#include <sys/types.h>
#include <ctype.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/err.h>
#include <stdio.h>
#include <string.h>

extern char *dns_txt(char *);

/* STARTHEAD */
/* Copyright Yahoo, Inc. 2004.
 * Redistribution beyond the Domain Signs List not permitted.
 * In time a proper open source license will available.
 */

#ifdef SWIG
%module domainkeys
%{
#include "domainkeys.h"
%}
#endif

typedef enum {
		      DK_STAT_OK, /* Function completed successfully */
		      DK_STAT_BADSIG, /* Signature was available but failed to verify against domain specified key */
		      DK_STAT_NOSIG, /* No signature available in message */
		      DK_STAT_NOKEY, /* No public key available (permanent failure) */
		      DK_STAT_BADKEY, /* Unusable key, public if verifying, private if signing */
		      DK_STAT_CANTVRFY, /* Cannot get domain key to verify signature (temporary failure) */
		      DK_STAT_SYNTAX, /* Message is not valid syntax. Signature could not be created/checked */
		      DK_STAT_NORESOURCE, /* Could not get critical resource (temporary failure) */
		      DK_STAT_ARGS, /* Arguments are not usable. */
		      DK_STAT_REVOKED,    /* Key has been revoked. */
		      DK_STAT_INTERNAL,	/* cannot call this routine in this context.  Internal error. */
		      } DK_STAT;

typedef enum {
  DK_FLAG_TESTING = 1,		/* set when in testing mode. */
  DK_FLAG_SIGNSALL = 2,		/* domain signs all outgoing email. */
  DK_FLAG_SET = 4,		/* flags set from a successful DNS query */
} DK_FLAGS;

typedef enum {
  DK_CANON_SIMPLE = 0,
  DK_CANON_NOFWS = 1,
} DK_CANON;

typedef struct {
/* STARTPRIV */
  const EVP_MD *md;
/* STOPPRIV */
} DK_LIB;

typedef struct {
/* STARTPRIV */
#define DKMARK ('D' | 'K'<<8 | 'E'<<16 | 'Y'<<24)
  int dkmarker;			/* in case somebody casts in */
  EVP_MD_CTX mdctx;		/* the hash */
#define DK_SIGNING_SIGN 0
#define DK_SIGNING_VERIFY 1
#define DK_SIGNING_NOSIGN 2
#define DK_SIGNING_NOVERIFY 3
  int signing;			/* our current signing/verifying state */
  int in_headers;		/* true if we're still processing headers */
  char *header;			/* points to a malloc'ed block for header. */
  int headerlen;		/* total length of the header. */
  int headermax;		/* length of malloc'ed block */
  int headerlinelen;		/* current column of the header */
  int headerptr;		/* current offset of the beginning of this header */
  char *from;			/* saved copy of From: header */
  char *sender;			/* saved copy of Sender: header */
  char *dksign;			/* saved copy of DK-Sig: header */
  char *domain;			/* d= */
  char *selector;		/* s= */
  char *signature;		/* b= */
  int canon;			/* c= canonicalization being applied. */
  int state;			/* state variable for canonicalization */
  char *headers;		/* h= */
  int errline;			/* source line of most recent error */
  char *errfile;		/* source file of most recent error */
/* STOPPRIV */
} DK;

/* STOPHEAD */

#define DKERR(x) ((dk->errline=__LINE__),(dk->errfile=__FILE__),(x))

/* HEADER */
/* returns the source file from which an error was returned. */
char * dk_errfile(DK *dk)
{
  return dk->errfile;
}

/* HEADER */
/* returns the source line number from which an error was returned. */
int dk_errline(DK *dk)
{
  return dk->errline;
}

/* HEADER */
/* Per-process, one-time initialization
 * Returns library structure for subsequent dk_sign or dk_verify calls.
 * Consult statp before using.
 */
DK_LIB *dk_init(DK_STAT *statp)
{
  DK_LIB *dklib;

  dklib = malloc(sizeof(DK_LIB));
  if (!dklib) {
    if (statp) *statp = DK_STAT_NORESOURCE;
    return NULL;
  }
  dklib->md = EVP_sha1();
  if (!dklib->md) {
    if (statp) *statp = DK_STAT_INTERNAL;
    free(dklib);
    return NULL;
  }
  if (statp) *statp = DK_STAT_OK;
  return dklib;
}

/* start a new header */
static DK_STAT
dkinit_new_header(DK *dk)
{
  dk->headermax = 1024;
  dk->header = malloc(dk->headermax);
  if (!dk->header) return DKERR(DK_STAT_NORESOURCE);
  dk->headerlen = 0;
  dk->headerlinelen = 1;	/* always store the first char. */
  dk->in_headers = 1;
  dk->headerptr = 0;
  dk->from = NULL;
  dk->sender = NULL;
  dk->dksign = NULL;
  dk->canon = DK_CANON_SIMPLE;
  dk->state = 0;
  dk->headers = 0;
  return DKERR(DK_STAT_OK);
}

static DK_STAT
dkstore_char(DK *dk, char ch)
{
  if (dk->headerlen >= dk->headermax) {
    char *hp;
    hp = malloc(dk->headermax * 2 + 1024 + 1); /* leave room for null */
    if (!hp) return DKERR(DK_STAT_NORESOURCE);
    if (dk->headermax) {
      memcpy(hp, dk->header, dk->headerlen);
      free(dk->header);
    }
    dk->header = hp;
    dk->headermax = dk->headermax * 2 + 1024;
  }
  dk->header[dk->headerlen++] = ch;
  dk->headerlinelen++;
  return DKERR(DK_STAT_OK);
}

/* HEADER */
/* Per-message, may be threaded.
 * canon is one of DK_CANON_*.
 * Returns state structure for operation.  Consult statp before using.
 */
DK *dk_sign(DK_LIB *dklib, DK_STAT *statp, int canon)
{
  DK *dk;
  BIO *rbio;

  dk = malloc(sizeof(DK));
  if (!dk) {
    if (statp) *statp = DKERR(DK_STAT_NORESOURCE);
    return NULL;
  }
  dk->dkmarker = DKMARK;
  dk->signing = DK_SIGNING_SIGN;
  if (dkinit_new_header(dk) != DK_STAT_OK) {
    /* couldn't malloc a header. */
    free(dk);
    if (statp) *statp = DKERR(DK_STAT_NORESOURCE);
    return NULL;
  }
  dk->canon = canon; /* TC13-simple, TC13-nofws */
  EVP_SignInit(&dk->mdctx, dklib->md);

  if (statp) *statp = DKERR(DK_STAT_OK);
  return dk;
}

/* HEADER */
/* Per-message, may be threaded.
 * Returns state structure for operation.  Consult statp before using.
 * id is printable opaque, used for printing in messages.
 */
DK *dk_verify(DK_LIB *dklib,
	      DK_STAT *statp)
{
  DK *dk;

  dk = malloc(sizeof(DK));
  if (!dk) {
    if (statp) *statp = DKERR(DK_STAT_NORESOURCE);
    return NULL;
  }
  dk->dkmarker = DKMARK;
  dk->signing = DK_SIGNING_NOVERIFY; /* wait to verify until we see DK-Signature. */
  if (dkinit_new_header(dk) != DK_STAT_OK) {
    /* couldn't malloc a header. */
    free(dk);
    if (statp) *statp = DKERR(DK_STAT_NORESOURCE);
    return NULL;
  }
  EVP_VerifyInit(&dk->mdctx, dklib->md);

  if (statp) *statp = DKERR(DK_STAT_OK);
  return dk;
}

/* parse an rfc822 address */

static DK_STAT
dkparse822(char *address, unsigned int offset)
{
  int angle = 0;		/* true if it's in angle brackets */
  char *from, *to;
  int level = 1;

  to = address;
  from = address + offset;
  while (*from) {		/* nuke comments. */
    if (*from == ')') {
      return DK_STAT_SYNTAX;
    } else if (*from == '\\' && from[1]) {
      /* skip backslash-quoted characters */
      *to++ = *from++;
      *to++ = *from++;
    } else if (*from == '"') {
      /* we have to ignore parenthesis inside quotes. */
      level = 1;
      from++;
      while (level) {
	if (*from == '\0') return DK_STAT_SYNTAX;
	else if (*from == '"') --level;
        else if (*from == '\\' && from[1]) *to++ = *from++;
	*to++ = *from++;
      }
    } else if (*from == '(') {
      level = 1;
      from++;
      while (level) {
	if (*from == '\0') return DK_STAT_SYNTAX;
	else if (*from == '(') ++level;
	else if (*from == ')') --level;
        else if (*from == '\\' && from[1]) ++from;
	++from;
      }
    } else if (*from == ' ') {
      ++from;
    } else if (*from == '\t') {
      ++from;
    } else {
      *to++ = *from++;
    }
  }
  *to = '\0';
  from = address;
  to = NULL;
  while (*from) {
    if (*from == '@') {
      to = address;
      ++from;
      continue;
    } else if (*from == '\\' && from[1]) {
      if (to) *to++ = *from;
      ++from;
    } else if (*from == '<') {
      angle = 1;
    } else if (angle && *from == '>') {
      break;
    } else if (to && *from == ' ') {
      break;
    } else if (to && *from == ';') {
      break;
    } else if (to && *from == '\t') {
      break;
    }
    if (to) *to++ = *from;
    from++;
  }
  if (to) *to = '\0';
  else *address = '\0';
  return DK_STAT_OK;
}

/* Given a list of key=value; pairs, find all the keys found in letters.
 * Store the value in the corresponding entry in values[].
 * Modifies list to insert nulls in the place of the semicolons which terminate
 * each of the values except possibly the last.
 * You can only call dkparselist() once on a given list.
 * Caller must ensure that values[] is as large as strlen(letters)
 * According to the spec, a value can have embedded spaces in it, however none
 * of the existing values need spaces, so we always remove them.
 */

static DK_STAT
dkparselist(char *list, char *letters, char *values[])
{
  char key;
  int i;
  char *value;

  /* start with all args unset */
  for (i = 0; letters[i]; i++)
    values[i] = NULL;

  key = 0;
  while (*list) {
    if (*list == ' ') {
      list++;
    } else if (*list == '\t') {
      list++;
    } else if (*list == '\r') {
      list++;
    } else if (*list == '\n') {
      list++;
    } else if (*list == '=') {
      char *ws;

      ++list;
      value = list;
      ws = list;
      while (1) {
	/* copy up to null or semicolon, deleting whitespace as we go */
	*ws = *list;
        if (*list == '\t') {
	  /* ignore */
	} else if (*list == ' ') {
	  /* ignore */
	} else if (*list == '\r') {
	  /* ignore */
	} else if (*list == '\n') {
	  /* ignore */
	} else if (!*list) {
	  break;
	} else if (*list == ';') {
	  *ws = '\0';
	  list++;
	  break;
	} else {
	  ws++;
	}
	list++;
      }
      if (!key) return DK_STAT_SYNTAX; /* we didn't get a key. TC22 */
      /* if we find a matching letter, remember the value */
      for (i = 0; letters[i]; i++) {
	if (key == letters[i]) {
	  if (values[i]) return DK_STAT_SYNTAX; /* no duplicate keys. TC23 */
	  values[i] = value;
	}
      }
      key = 0;
    } else {
      if (key) return DK_STAT_SYNTAX; /* they already gave us a key. TC24 */
      key = *list++;
      if (!islower(key)) return DK_STAT_SYNTAX;	/* must be lowercase letter. TC25 */
    }
  }
  if (key) return DK_STAT_SYNTAX; /* we ended up with an extra key. TC25-1 */
  return DK_STAT_OK;
}

static DK_FLAGS
dkpolicy(DK *dk)
{
  char *query, *results, *flags[2];
  char *domain = NULL;
  int dkf = 0;

  if (dk->dksign) domain = dk->domain;
  if (!domain) domain = dk->sender;
  if (!domain) domain = dk->from;
  if (!domain) return 0;

  query = malloc(strlen("_domainkey.") + strlen(domain) + 1);
  if (query) {
    sprintf(query, "_domainkey.%s", domain);
    results = dns_txt(query);
    free(query);
    
    if (!strcmp(results,"e=perm;"))
      ;
    else if (!strcmp(results,"e=temp;"))
      ;
    else {
      dkparselist(results, "ot", flags);
      
      if (flags[0] && *flags[0] == '-')  /* TC36 */
	dkf |= DK_FLAG_SIGNSALL;
      if (flags[1] && *flags[1] == 'y')  /* TC36-1 */
	dkf |= DK_FLAG_TESTING;
      dkf |= DK_FLAG_SET;
      free(results);
    }
  }
  return dkf;
}

static
dkhash(DK *dk, const unsigned char *ptr)
{
  if (dk->canon == DK_CANON_NOFWS && (*ptr == ' ' || *ptr == '\t'))
    return;
  switch(dk->signing) {
  case DK_SIGNING_SIGN:
  case DK_SIGNING_VERIFY:
    if (*ptr == '\r' && ((dk->state & 1) == 0)) dk->state++;
    else if (*ptr == '\n' && ((dk->state & 1) == 1)) dk->state++;
    else {
      while (dk->state >= 2) {
	EVP_DigestUpdate(&dk->mdctx, "\r\n", 2);
	//printf("\r\n");
	dk->state -= 2;
      }
      if (dk->state) {
	EVP_DigestUpdate(&dk->mdctx, "\r", 1);
        //printf("\r");
	dk->state --;
      }
      EVP_DigestUpdate(&dk->mdctx, ptr, 1);
      //printf("%c", *ptr);
    }
  }
}

/* process headers if there's a h=*/
/* iii is the offset into headers pointing at the first header after DK-Sig */
static DK_STAT
dkheaders_header(DK *dk, int iii)
{
  int i, j, k, m;

  /* Yes, this is an O(N*M) algorithm where N is the headers listed in h= and
   * M is the number of headers in the email.  N is pretty small -- O(10), so it's likely
   * that anything which requires pre-computation won't help. */
  j = 0;
  i = 0;
  for (;;) {
    if (!dk->headers[i] || dk->headers[i] == ':') {
      m = iii;
      for (k = iii; k < dk->headerlen; k++) if (dk->header[k] == '\0') {
	if (!strncasecmp(dk->header + m, dk->headers + j, i - j)) {
	  while (m < k) {
	    dkhash(dk, dk->header + m++);
	  }
	  dkhash(dk, "\r");
	  dkhash(dk, "\n");
	}
	m = k + 1;
      }
      j = i + 1;
    }
    if (!dk->headers[i]) break;
    i++;
  }
  return DKERR(DK_STAT_OK);
}

/* process headers */
static DK_STAT
dkheaders(DK *dk)
{
  int i;
  int start_signed;

  for (i = 0; i < dk->headerlen; i++) {
    if (dk->headers && dk->signing == DK_SIGNING_VERIFY) ;
    else if (dk->canon == DK_CANON_NOFWS && (dk->header[i] == '\r' || dk->header[i] == '\n')) ;
    else if (dk->header[i]) dkhash(dk, dk->header + i);
    else {dkhash(dk, "\r");dkhash(dk, "\n");}

    if (i==0 || dk->header[i-1] == '\0') {
      if (!strncasecmp(dk->header + i, "From:", 5)) {
	/* Remember the From: and forget the current. */
	if (dk->from) {
	  /* if we already got a From: header, fuhgeddabout it. */
	return DKERR(DK_STAT_SYNTAX);
	}
	dk->from = strdup(dk->header + i);
	if (dkparse822(dk->from, 5) != DK_STAT_OK)
	  return DKERR(DK_STAT_SYNTAX);
      } else if (!strncasecmp(dk->header + i, "Sender:", 7)) {
	if (dk->sender) {
	  /* if we already got a Sender: header, fuhgeddabout it. */
	  return DKERR(DK_STAT_SYNTAX);
	}
	if (dk->signing != DK_SIGNING_NOVERIFY) {
	  /* only remember the Sender if we're verifying already. */
	  dk->sender = strdup(dk->header + i);
	  if (dkparse822(dk->sender, 7) != DK_STAT_OK)
	    return DKERR(DK_STAT_SYNTAX);
	}
      } else if (!strncasecmp(dk->header + i, "DomainKey-Signature:", 20)) {
	int thisheaderlen = strlen(dk->header + i);

	/* Do not sign email that already has a dksign unless the Sender was found first, 3.5.2 TC41, TC41-1*/
	if (dk->signing == DK_SIGNING_SIGN && !dk->sender)
	  return DKERR(DK_STAT_SYNTAX);
	/* check the outermost (first encountered) signature */
	if (!dk->dksign) {
	  char *values[7];	/* dsbchqa */
	  
	  dk->dksign = dk->header + i;
	    
	  /* parse the dksign header .*/
	  if (dkparselist(dk->dksign + 20,	/* magic number: length of header name */
			  "dsbchqa", values) != DK_STAT_OK)
	    return DKERR(DK_STAT_SYNTAX);
	    
	  dk->domain = values[0];
	  dk->selector = values[1];
	  dk->signature = values[2];
	  if (!dk->selector || !dk->domain || !dk->signature) {
	    /* we really do need to have a domain, selector and key. TC21 TC40 */
	    return DKERR(DK_STAT_NOSIG);
	  }

	  if (!values[3])
	    dk->canon = DK_CANON_SIMPLE;
	  else if (!strcasecmp(values[3], "simple"))
	    dk->canon = DK_CANON_SIMPLE;
	  else if (!strcasecmp(values[3], "nofws"))
	    dk->canon = DK_CANON_NOFWS;
	  else return DKERR(DK_STAT_SYNTAX); /* TC42 */

	  dk->headers = values[4];

	  if (values[5] && strcasecmp(values[5], "dns")) /* TC42-2 */
	    return DKERR(DK_STAT_SYNTAX);
	  if (values[6] && strcasecmp(values[6], "rsa-sha1"))	/* TC42-3 */
	    return DKERR(DK_STAT_SYNTAX);

	}
	/* if we're waiting to verify, start now. */
	if (dk->signing == DK_SIGNING_NOVERIFY) {
	  dk->signing = DK_SIGNING_VERIFY; /* the signature starts here */
	}
	i += thisheaderlen;
	start_signed = i;
      } 
    }
  }
  if (dk->signing == DK_SIGNING_NOVERIFY)
    /* No DK-Sig: should return No Signature. */
    return DKERR(DK_STAT_NOSIG);
  if (!dk->from) {
    /* No From:, 3.1 says that it's no good..  TC11/TC16 */
    return DKERR(DK_STAT_SYNTAX);
  }
  if (dk->headers && dk->signing == DK_SIGNING_VERIFY)
    return dkheaders_header(dk, start_signed);
  return DKERR(DK_STAT_OK);
}


/* HEADER */
/* Must NOT include dots inserted for SMTP encapsulation.
 * Must NOT include CRLF.CRLF which terminates the message.
 * Otherwise must be exactly that which is sent or received over the SMTP session.
 * May be called multiple times (not necessary to read an entire message into memory).
 */
DK_STAT   dk_message(DK *dk, const unsigned char *ptr, size_t len)
{
  DK_STAT st;

  if (!dk) return DK_STAT_ARGS;
  if (dk->dkmarker != DKMARK) return DK_STAT_ARGS;
  if (len && !ptr) return DKERR(DK_STAT_ARGS);

  while (len--) {
    /* parse headers */
    if (!dk->in_headers) {
      dkhash(dk, ptr);
    } else if (*ptr == '\n' && dk->headerlinelen != 0) {
      /* beginning a new line, but we can't do anything until we get a char. */
      dk->headerlinelen = 0;
    } else if (*ptr == '\r') {
      /* ignore carriage-returns, even bare ones.  We'll add them back in later.*/
    } else if (dk->headerlinelen) {
      /* if we're not starting a new header, just store it */
      if (dkstore_char(dk, *ptr) != DK_STAT_OK)
	return DKERR(DK_STAT_NORESOURCE);
    } else if (*ptr == ' ' || *ptr == '\t') {
      /* a new header ... starting with whitespace.  Must be continuation */
      /* just remember the whitespace */
      if (dkstore_char(dk, *ptr) != DK_STAT_OK)
	return DKERR(DK_STAT_NORESOURCE);
    } else if (*ptr == '\n') {
      /* an empty line terminates header processing */
      /* terminate the previous header */
      if (dkstore_char(dk, '\0') != DK_STAT_OK)
	return DKERR(DK_STAT_NORESOURCE);
      dk->in_headers = 0;
      st = dkheaders(dk);
      dkhash(dk, "\r");
      dkhash(dk, "\n");
      if (st != DK_STAT_OK) return st;
    } else {
      /* we're starting a new header.  Terminate the previous one. */
      if (dkstore_char(dk, '\0') != DK_STAT_OK)
	return DKERR(DK_STAT_NORESOURCE);
      /* remember the first character of a new header. */
      if (dkstore_char(dk, *ptr) != DK_STAT_OK)
	return DKERR(DK_STAT_NORESOURCE);
    }
    ptr++;
  }
  return DKERR(DK_STAT_OK);
}

/* HEADER */
/* Returns a pointer to a null-terminated domain name portion of an RFC 2822 address.
 * If a Sender: was encountered, it returns that domain.  Otherwise,
 * if a From: was encountered, it returns that domain.  Otherwise,
 * return NULL.
 * return NULL if no domain name found in the address.
 * return NULL if the dk is unusable for any reason.
 * return NULL if the address is unusable for any reason.
 */
char *
dk_from(DK *dk)
{
  if (!dk) return NULL;
  if (dk->dkmarker != DKMARK) return NULL;

  if (dk->sender) return dk->sender;
  if (dk->from) return dk->from;
  return NULL;
}

/* HEADER */
/*
 * Called at end-of-message (before response to DATA-dot, if synchronous with SMTP session).
 * If verifying, returns signature validity.
 * This does not calculate the signature.  Call dk_getsig() for that.
 * Flags are returned indirectly through dkf.
 * If you pass in NULL for dkf, the flags will not be fetched.
 * If there is a DK-Sig line, the d= entry will be used to fetch the flags.
 * Otherwise the Sender: domain will be used to fetch the flags.
 * Otherwise the From: domain will be used to fetch the flags.
 */
DK_STAT   dk_eom(DK *dk, DK_FLAGS *dkf)
{
  if (!dk) return DK_STAT_ARGS;
  if (dk->dkmarker != DKMARK) return DK_STAT_ARGS;

#if 0
  /* add another blank line to make sure that the headers are
   * recognized.  technically, this is a hack, since we're relying on
   * the trailing CRLF triggering the end of header recognition AND
   * also being deleted by the canonicalization algorithm.
   */
  dk_message(dk, "\r\n", 2);
#else
  if (dk->in_headers) {
    DK_STAT st;

    if (dkstore_char(dk, '\0') != DK_STAT_OK)
      return DKERR(DK_STAT_NORESOURCE);
    st = dkheaders(dk);
    if (st != DK_STAT_OK) return st;
  }
#endif
  
  switch (dk->signing) {
  case DK_SIGNING_SIGN:
  case DK_SIGNING_VERIFY:
    if (dk->state)
      EVP_DigestUpdate(&dk->mdctx, "\r\n", 2);
  }

  switch (dk->signing) {
  case DK_SIGNING_SIGN:
  case DK_SIGNING_NOSIGN:
    if (!dk->from) return DKERR(DK_STAT_SYNTAX);
    return DKERR(DK_STAT_OK);
  case DK_SIGNING_VERIFY:
  case DK_SIGNING_NOVERIFY: {
    unsigned char md_value[1024];
    int md_len, i;
    char *s;
    int size;
    char *domainkeys; /* malloc'ed */
    char *txtrec; /* malloc'ed */
    char *pubkeyvals[3];
    char pkey[4096]; /* XXX: acckkkpthththththt */
    int pkeylen;
    int okay;
    char *errorstr;
    DK_STAT st;

    unsigned int siglen;
    unsigned char *sig;
    BIO *bio, *b64;
    EVP_PKEY *publickey;
    unsigned long errornum;

    if (dkf) *dkf = 0;		/* clear flags first. */

    /* make sure that we got a header */
    if (!dk->dksign) {
      if (dkf) *dkf = dkpolicy(dk);
      return DKERR(DK_STAT_NOSIG); /* TC11 */
    }

    /* make sure that domain on the From: matches the d data. TC26 */
    /* Or is a subdomain of the d data.  TC26-1, etc. */
    if (dk->sender) s = dk->sender;
    else s = dk->from;
    if (!s)
      return DKERR(DK_STAT_SYNTAX);
    i = 0;
    if (strcasecmp(dk->domain, s))
      for (; i < strlen(s); i++)
	if (s[i] == '.')
	  if (!strcasecmp(dk->domain, s + i + 1))
	    break;
    if (i == strlen(s))
      return DKERR(DK_STAT_SYNTAX);

    /* convert their signature from base64 into binary */
    bio = BIO_new_mem_buf(dk->signature, -1);
    b64 = BIO_new(BIO_f_base64());
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    BIO_push(b64, bio);
    md_len = BIO_read(b64, md_value, sizeof(md_value));
    BIO_free_all(bio);
    if (md_len >= sizeof(md_value)) return DKERR(DK_STAT_NORESOURCE);

    /* get the s data and the d data and lookup s._domainkey.d */
    domainkeys = malloc(strlen(dk->selector) + strlen(dk->domain) + strlen("._domainkey.") + 1);
    if (!domainkeys) return DKERR(DK_STAT_NORESOURCE);
    sprintf(domainkeys, "%s._domainkey.%s", dk->selector, dk->domain);
    txtrec = dns_txt(domainkeys);
    free(domainkeys);

    if (!strcmp(txtrec,"e=perm;")) /* TC31 */
      return DKERR(DK_STAT_NOKEY);

    if (!strcmp(txtrec,"e=temp;")) /* TC30 */
      return DKERR(DK_STAT_CANTVRFY);

    if (dkparselist(txtrec, "pto", pubkeyvals) != DK_STAT_OK) {
      free(txtrec);
      return DKERR(DK_STAT_BADKEY);
    }

    if (!pubkeyvals[0]) {
      free(txtrec);
      return DKERR(DK_STAT_NOKEY); /* TC27 */
    }

    if (!*pubkeyvals[0]) {
      free(txtrec);
      return DKERR(DK_STAT_REVOKED); /* TC27 */
    }

    if (dkf) {
      /* TC35 and TC37 */
      if (pubkeyvals[1] && *pubkeyvals[1] == 'y')
	*dkf |= DK_FLAG_TESTING;
    }

    /* convert their public key from base64 into binary */
    bio = BIO_new_mem_buf(pubkeyvals[0], -1);
    if (!bio) {
      free(txtrec);
      return DKERR(DK_STAT_NORESOURCE);
    }
    b64 = BIO_new(BIO_f_base64());
    if (!b64) {
      free(txtrec);
      BIO_free(bio);
      return DKERR(DK_STAT_NORESOURCE);
    }
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    BIO_push(b64, bio);
    publickey = d2i_PUBKEY_bio(b64, NULL);
    BIO_free_all(bio);
    free(txtrec);
    if (!publickey) return DKERR(DK_STAT_BADKEY);

    /* using that key, verify that the digest is properly signed */
    okay = EVP_VerifyFinal(&dk->mdctx, md_value, md_len, publickey);
    
    if (okay > 0) {
      st = DK_STAT_OK;
    } else {
      if (dkf) *dkf = dkpolicy(dk);
      st = DK_STAT_BADSIG;
    }
    EVP_PKEY_free(publickey);
    return DKERR(st);
  }
  }
}


/* HEADER */
/*
 * 
 * privatekey is the private key used to create the signature; It should contain
 * the entire contents of a PEM-format private key file, thusly it will begin with
 * -----BEGIN RSA PRIVATE KEY-----.  It should be null-terminated.
 */
size_t dk_siglen(void *privatekey)
{
  BIO *bio;
  EVP_PKEY *pkey;
  size_t len;

  if (!privatekey) return 0;

  bio = BIO_new_mem_buf(privatekey, -1);
  pkey = PEM_read_bio_PrivateKey(bio, NULL, NULL, NULL);
  BIO_free(bio);

  len = (EVP_PKEY_size(pkey) + 2) / 3 * 4;
  EVP_PKEY_free(pkey);
  return len;
}


/* HEADER */
/*
 * Sets buf to a null-terminated string.
 * If the message is being signed, signature is stored in the buffer.
 * If the message is being verified, returns DK_STAT_INTERNAL.
 * privatekey is the private key used to create the signature; It should contain
 * the entire contents of a PEM-format private key file, thusly it will begin with
 * -----BEGIN RSA PRIVATE KEY-----.  It should be null-terminated.
 * If you pass in NULL for buf, you'll get back DK_STAT_NORESOURCE.
 * If len is not big enough, you'll get back DK_STAT_NORESOURCE.
 */
DK_STAT   dk_getsig(DK *dk, void *privatekey, unsigned char buf[], size_t len)
{
  if (!dk) return DK_STAT_ARGS;
  if (dk->dkmarker != DKMARK) return DK_STAT_ARGS;
  if (!privatekey) return DK_STAT_ARGS;
  if (!buf) return DKERR(DK_STAT_NORESOURCE);
  switch (dk->signing) {
  case DK_SIGNING_SIGN:
  case DK_SIGNING_NOSIGN: {
    unsigned int siglen;
    unsigned char *sig;
    int size;
    BIO *bio, *b64;
    EVP_PKEY *pkey;

    bio = BIO_new_mem_buf(privatekey, -1);
    pkey = PEM_read_bio_PrivateKey(bio, NULL, NULL, NULL);
    BIO_free(bio);

    if (!pkey) {
      /* their private key is no good TC33 */
      return DKERR(DK_STAT_BADKEY);
    }

    siglen = EVP_PKEY_size(pkey);
    sig = (unsigned char*) OPENSSL_malloc(siglen);
    EVP_SignFinal(&dk->mdctx, sig, &siglen, pkey);
    EVP_PKEY_free(pkey);

    bio = BIO_new(BIO_s_mem());
    if (!bio) {
      return DKERR(DK_STAT_NORESOURCE);
    }
    b64 = BIO_new(BIO_f_base64());
    if (!b64) {
      BIO_free(bio);
      return DKERR(DK_STAT_NORESOURCE);
    }
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    BIO_push(b64, bio);
    if (BIO_write(b64, sig, siglen) < siglen) {
      OPENSSL_free(sig);
      BIO_free_all(bio);
      return DKERR(DK_STAT_NORESOURCE);
    }
    BIO_flush(b64);
    OPENSSL_free(sig);

    size = BIO_read(bio, buf, len);
    BIO_free_all(bio);

    if (size >= len) return DKERR(DK_STAT_NORESOURCE); /* TC28 */

    buf[size] = '\0';
    return DKERR(DK_STAT_OK);
  }
  case DK_SIGNING_VERIFY:
  case DK_SIGNING_NOVERIFY:
    return DKERR(DK_STAT_INTERNAL);
  }
}


/* HEADER */
/*
 * Free all resources associated with this message.
 * dk is no longer usable.
 */
DK_STAT   dk_free(DK *dk)
{
  if (!dk) return DK_STAT_ARGS;
  if (dk->dkmarker != DKMARK) return DK_STAT_ARGS;
  if (dk->from) free(dk->from);
  if (dk->sender) free(dk->from);
  free(dk->header);		/* alloc'ing dk->header is not optional. */
  dk->dkmarker = ~DKMARK;
  free(dk);
  return DK_STAT_OK;
}

static char *errors[] = {
  "DK_STAT_OK: Function completed successfully",
  "DK_STAT_BADSIG: Signature was available but failed to verify against domain specified key",
  "DK_STAT_NOSIG: No signature available in message",
  "DK_STAT_NOKEY: No public key available (permanent failure)",
  "DK_STAT_BADKEY: Unusable key, public if verifying, private if signing.",
  "DK_STAT_CANTVRFY: Cannot get domain key to verify signature (temporary failure)",
  "DK_STAT_SYNTAX: Message is not valid syntax. Signature could not be created/checked",
  "DK_STAT_NORESOURCE: Could not get critical resource (temporary failure)",
  "DK_STAT_ARGS: Arguments are not usable.",
  "DK_STAT_REVOKED: Key has been revoked.",
  "DK_STAT_INTERNAL: cannot call this routine in this context.  Internal error.",
};

/* HEADER */
/*
 * return a pointer to a string which describes st.
 * The string is structured.  All the characters up to the first colon
 * contain the name of the DK_STAT constant.  From there to the end of
 * string is a human-readable description of the error.
 */
const char *DK_STAT_to_string(DK_STAT st)
{
  if (st >= (sizeof errors) / (sizeof errors[0]))
    return "DK_STAT_UNKNOWN: unknown status";
  else
    return errors[st];
}
